/*
 * GQview
 * (C) 2003 John Ellis
 *
 * Author: John Ellis
 *
 * This software is released under the GNU General Public License (GNU GPL).
 * Please read the included file COPYING for more information.
 * This software comes with no warranty of any kind, use at your own risk!
 */


#include "gqview.h"
#include "bar_exif.h"

#include "exif.h"

#include <math.h>

#define EXIF_BAR_SIZE_INCREMENT 32
#define EXIF_BAR_ARROW_SIZE 7


/*
 *-------------------------------------------------------------------
 * table util
 *-------------------------------------------------------------------
 */

static GtkWidget *table_add_line(GtkWidget *table, gint x, gint y,
				 const gchar *description, const gchar *text)
{
	GtkWidget *label;

	if (!text) text = "";

	label = gtk_label_new(description);
	gtk_misc_set_alignment(GTK_MISC(label), 1.0, 0.5);
	gtk_table_attach(GTK_TABLE(table), label,
			 x, x + 1, y, y + 1,
			 GTK_FILL, 0,
			 2, 2);
	gtk_widget_show(label);

	label = gtk_label_new(text);
	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
	gtk_table_attach(GTK_TABLE(table), label,
			 x + 1, x + 2, y, y + 1,
			 GTK_FILL, 0,
			 2, 2);
	gtk_widget_show(label);
	return label;
}

/*
 *-------------------------------------------------------------------
 * EXIF bar
 *-------------------------------------------------------------------
 */

typedef struct _ExifBar ExifBar;
struct _ExifBar
{
	GtkWidget *vbox;
	GtkWidget *table;
	GtkWidget *label_camera;
	GtkWidget *label_date;
	GtkWidget *label_aperture;
	GtkWidget *label_exposure_program;
	GtkWidget *label_exposure_bias;
	GtkWidget *label_focal_length;
	GtkWidget *label_metering_mode;
	GtkWidget *label_subject_distance;
	GtkWidget *label_flash;
	GtkWidget *label_light_source;
	GtkWidget *label_shutter_speed;
	GtkWidget *label_iso;
	GtkWidget *label_resolution;
	GtkWidget *label_orientation;
};

static void bar_exif_update(ExifBar *eb, const gchar *path)
{
	ExifData *exif;
	gchar *buf;
	const gchar *text;

	if (!path)
		{
		gtk_widget_set_sensitive(eb->table, FALSE);
		return;
		}

	exif = exif_read(path);

	if (!exif)
		{
		gtk_widget_set_sensitive(eb->table, FALSE);
		return;
		}

	gtk_widget_set_sensitive(eb->table, TRUE);

	if (exif->make && exif->model)
		{
		text = " ";
		}
	else
		{
		text = "";
		}
	buf = g_strdup_printf("%s%s%s%s%s%s", (exif->make) ? exif->make : "", text,
					    (exif->model) ? exif->model : "",
					    (exif->software) ? " (" : "",
					    (exif->software) ? exif->software : "",
					    (exif->software) ? ")" : "");
	gtk_label_set(GTK_LABEL(eb->label_camera), buf);
	g_free(buf);

	gtk_label_set(GTK_LABEL(eb->label_date), (exif->date_original) ? exif->date_original : "");

	buf = g_strdup_printf("f/%.1f", exif->fnumber != 0.0 ? exif->fnumber : exif->aperture);
	gtk_label_set(GTK_LABEL(eb->label_aperture), buf);
	g_free(buf);

	gtk_label_set(GTK_LABEL(eb->label_exposure_program),
		      exif_get_info_string(exif, EXIF_EXPOSURE_PROGRAM));

	buf = g_strdup_printf("%+.1f", exif->exposure_bias);
	gtk_label_set(GTK_LABEL(eb->label_exposure_bias), buf);
	g_free(buf);

	buf = g_strdup_printf("%.2f mm", exif->focal_length);
	gtk_label_set(GTK_LABEL(eb->label_focal_length), buf);
	g_free(buf);

	if (exif->subject_distance != 0.0)
		{
		buf = g_strdup_printf("%.3f m", exif->subject_distance);
		}
	else
		{
		buf = g_strdup(_("unknown"));
		}
	gtk_label_set(GTK_LABEL(eb->label_subject_distance), buf);
	g_free(buf);

	gtk_label_set(GTK_LABEL(eb->label_metering_mode),
		      exif_get_info_string(exif, EXIF_METERING_MODE));

	gtk_label_set(GTK_LABEL(eb->label_flash), 
		      exif_get_info_string(exif, EXIF_FLASH));
	gtk_label_set(GTK_LABEL(eb->label_light_source),
		      exif_get_info_string(exif, EXIF_LIGHT_SOURCE));

	if (exif->shutter_speed != 0.0)
		{
		buf = g_strdup_printf("%s%.0fs", exif->shutter_speed > 1.0 ? "1/" : "",
						 exif->shutter_speed > 1.0 ? floor(exif->shutter_speed) : 1.0 / exif->shutter_speed);
		}
	else if (exif->exposure_time.den != 0 && exif->exposure_time.num != 0)
		{
		double exp;

		/* exposure time as 1/exp seconds */
		exp = (double)exif->exposure_time.den / (double)exif->exposure_time.num;

		/* display as "1/60s" or just "8s" */
		buf = g_strdup_printf("%s%.0fs", exp > 1 ? "1/" : "", exp > 1 ? exp : 1 / exp);
		}
	else
		{
		buf = g_strdup(_("unknown"));
		}
	gtk_label_set(GTK_LABEL(eb->label_shutter_speed), buf);
	g_free(buf);

	if (exif->iso_sensivity != 0)
		{
		buf = g_strdup_printf("%d", exif->iso_sensivity);
		}
	else
		{
		buf = g_strdup(_("unknown"));
		}
	gtk_label_set(GTK_LABEL(eb->label_iso), buf);
	g_free(buf);

	buf = exif_orientation_string(exif);
	gtk_label_set(GTK_LABEL(eb->label_orientation), buf);
	g_free(buf);

	text = exif_get_info_string(exif, EXIF_RESUNIT);
	if (!text || strlen(text) < 1) text = _("unknown");
	buf = g_strdup_printf("%0.f x %0.f (%s/%s)", exif->resolution_x.den ? (double)exif->resolution_x.num / (exif->resolution_x.den) : 1.0,
						  exif->resolution_y.den ? (double)exif->resolution_y.num / (exif->resolution_y.den) : 1.0,
						  _("dot"),
						  text);
	gtk_label_set(GTK_LABEL(eb->label_resolution), buf);
	g_free(buf);

	exif_free(exif);
}

static void bar_exif_clear(ExifBar *eb)
{
	if (!GTK_WIDGET_SENSITIVE(eb->label_camera)) return;
	
	gtk_label_set(GTK_LABEL(eb->label_camera), "");
	gtk_label_set(GTK_LABEL(eb->label_date), "");
	gtk_label_set(GTK_LABEL(eb->label_aperture), "");
	gtk_label_set(GTK_LABEL(eb->label_exposure_program), "");
	gtk_label_set(GTK_LABEL(eb->label_exposure_bias), "");
	gtk_label_set(GTK_LABEL(eb->label_focal_length), "");
	gtk_label_set(GTK_LABEL(eb->label_subject_distance), "");
	gtk_label_set(GTK_LABEL(eb->label_metering_mode), "");
	gtk_label_set(GTK_LABEL(eb->label_flash), "");
	gtk_label_set(GTK_LABEL(eb->label_light_source), "");
	gtk_label_set(GTK_LABEL(eb->label_shutter_speed), "");
	gtk_label_set(GTK_LABEL(eb->label_iso), "");
	gtk_label_set(GTK_LABEL(eb->label_resolution), "");
	gtk_label_set(GTK_LABEL(eb->label_orientation), "");
}

void bar_exif_set(GtkWidget *bar, const gchar *path)
{
	ExifBar *eb;

	eb = g_object_get_data(G_OBJECT(bar), "bar_exif_data");
	if (!eb) return;

	bar_exif_clear(eb);
	bar_exif_update(eb, path);
}

void bar_exif_close(GtkWidget *bar)
{
	ExifBar *eb;

	eb = g_object_get_data(G_OBJECT(bar), "bar_exif_data");
	if (!eb) return;

	gtk_widget_destroy(eb->vbox);
}

static void bar_exif_size(ExifBar *eb, gint val)
{
	gint size;

	size = eb->vbox->allocation.width;
	size = CLAMP(size + val, EXIF_BAR_SIZE_INCREMENT * 2, EXIF_BAR_SIZE_INCREMENT * 16);

	gtk_widget_set_size_request(eb->vbox, size, -1);
}

static void bar_exif_larger(GtkWidget *widget, gpointer data)
{
	ExifBar *eb = data;

	bar_exif_size(eb, EXIF_BAR_SIZE_INCREMENT);
}

static void bar_exif_smaller(GtkWidget *widget, gpointer data)
{
	ExifBar *eb = data;

	bar_exif_size(eb, -EXIF_BAR_SIZE_INCREMENT);
}

static void bar_exif_destroy(GtkWidget *widget, gpointer data)
{
	ExifBar *eb = data;

	g_free(eb);
}

GtkWidget *bar_exif_new(gint show_title, const gchar *path)
{
	ExifBar *eb;
	GtkWidget *table;
	GtkWidget *scrolled;

	eb = g_new0(ExifBar, 1);

	eb->vbox = gtk_vbox_new(FALSE, 0);
	g_object_set_data(G_OBJECT(eb->vbox), "bar_exif_data", eb);
	g_signal_connect(G_OBJECT(eb->vbox), "destroy",
			 G_CALLBACK(bar_exif_destroy), eb);

	if (show_title)
		{
		GtkWidget *box;
		GtkWidget *label;
		GtkWidget *button;
		GtkWidget *arrow;

		box = gtk_hbox_new(FALSE, 0);

		label = gtk_label_new(_("Exif"));
		gtk_box_pack_start(GTK_BOX(box), label, TRUE, TRUE, 0);
		gtk_widget_show(label);

		button = gtk_button_new();
		gtk_button_set_relief(GTK_BUTTON(button), GTK_RELIEF_NONE);
		g_signal_connect(G_OBJECT(button), "clicked",
				 G_CALLBACK(bar_exif_smaller), eb);
		gtk_box_pack_end(GTK_BOX(box), button, FALSE, FALSE, 0);
		arrow = gtk_arrow_new(GTK_ARROW_RIGHT, GTK_SHADOW_NONE);
		gtk_widget_set_size_request(arrow, EXIF_BAR_ARROW_SIZE, EXIF_BAR_ARROW_SIZE);
		gtk_container_add(GTK_CONTAINER(button), arrow);
		gtk_widget_show(arrow);
		gtk_widget_show(button);

		button = gtk_button_new();
		gtk_button_set_relief(GTK_BUTTON(button), GTK_RELIEF_NONE);
		g_signal_connect(G_OBJECT(button), "clicked",
				 G_CALLBACK(bar_exif_larger), eb);
		gtk_box_pack_end(GTK_BOX(box), button, FALSE, FALSE, 0);
		arrow = gtk_arrow_new(GTK_ARROW_LEFT, GTK_SHADOW_NONE);
		gtk_widget_set_size_request(arrow, EXIF_BAR_ARROW_SIZE, EXIF_BAR_ARROW_SIZE);
		gtk_container_add(GTK_CONTAINER(button), arrow);
		gtk_widget_show(arrow);
		gtk_widget_show(button);

		gtk_box_pack_start(GTK_BOX(eb->vbox), box, FALSE, FALSE, 0);
		gtk_widget_show(box);
		}

	table = gtk_table_new(2, 9, FALSE);
	gtk_container_set_border_width(GTK_CONTAINER(table), 5);

	eb->table = table;
	eb->label_camera = table_add_line(table, 0, 0, _("Camera:"), NULL);
	eb->label_date = table_add_line(table, 0, 1, _("Date:"), NULL);
	eb->label_shutter_speed = table_add_line(table, 0, 2, _("Shutter speed:"), NULL);
	eb->label_aperture = table_add_line(table, 0, 3, _("Aperture:"), NULL);
	eb->label_exposure_program = table_add_line(table, 0, 4, _("Exposure program:"), NULL);
	eb->label_exposure_bias = table_add_line(table, 0, 5, _("Exposure bias:"), NULL);
	eb->label_iso = table_add_line(table, 0, 6, _("ISO sensitivity:"), NULL);
	eb->label_focal_length = table_add_line(table, 0, 7, _("Focal length:"), NULL);
	eb->label_subject_distance = table_add_line(table, 0, 8, _("Subject distance:"), NULL);
	eb->label_metering_mode = table_add_line(table, 0, 9, _("Metering mode:"), NULL);
	eb->label_flash = table_add_line(table, 0, 10, _("Flash:"), NULL);
	eb->label_light_source = table_add_line(table, 0, 11, _("Light source:"), NULL);
	eb->label_resolution = table_add_line(table, 0, 12, _("Resolution:"), NULL);
	eb->label_orientation = table_add_line(table, 0, 13, _("Orientation:"), NULL);

	scrolled = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(scrolled), table);
	gtk_widget_show(table);

	gtk_box_pack_start(GTK_BOX(eb->vbox), scrolled, TRUE, TRUE, 0);
	gtk_widget_show(scrolled);

	bar_exif_update(eb, path);

	return eb->vbox;
}

