/*
 *  GQView
 *  (C) 2003 John Ellis
 *
 *  Authors:
 *    Support for Exif file format, written by Eric Swalens.    
 *    Modified by Quy Tonthat
 *    

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <inttypes.h>	/* stdint.h is not available on all systems... */
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <math.h>

#include <glib.h>

#include "intl.h"

#include "exif.h"

#define BYTE_ORDER_INTEL	1
#define BYTE_ORDER_MOTOROLA	2

#define MARKER_UNKNOWN		0x00
#define MARKER_SOI		0xD8
#define MARKER_APP1		0xE1

#define FORMAT_UBYTE		1
#define FORMAT_STRING		2
#define	FORMAT_USHORT		3
#define	FORMAT_ULONG		4
#define	FORMAT_URATIONAL	5
#define	FORMAT_SBYTE		6
#define	FORMAT_UNDEFINED	7
#define	FORMAT_SSHORT		8
#define	FORMAT_SLONG		9
#define	FORMAT_SRATIONAL	10
#define	FORMAT_FLOAT		11
#define	FORMAT_DOUBLE		12

int IFD_format_size[] = {1, 1, 2, 4, 8, 1, 1, 2, 4, 8, 4, 8};

#define TAG_IMAGE_DESC		0x10E
#define TAG_MAKE		0x10F
#define TAG_MODEL		0x110
#define TAG_ORIENTATION		0x112
#define TAG_RESOLUTION_X	0x11a
#define TAG_RESOLUTION_Y	0x11b
#define TAG_RESOLUTION_UNIT	0x128
#define TAG_SOFTWARE		0x131
#define TAG_DATETIME		0x132
#define TAG_COPYRIGHT		0x8298
#define TAG_EXIFOFFSET		0x8769
#define TAG_EXPOSURE_TIME	0x829A
#define TAG_FNUMBER		0x829D
#define TAG_EXPOSURE_PROGRAM	0x8822
#define TAG_ISO_SPEED		0x8827
#define TAG_DATETIME_ORIGINAL	0x9003
#define TAG_DATETIME_DIGITIZED	0x9004
#define TAG_SHUTTER_SPEED	0x9201
#define TAG_APERTURE		0x9202
#define TAG_EXPOSURE_BIAS	0x9204
#define TAG_APERTURE_MAX	0x9205
#define TAG_SUBJECT_DISTANCE	0x9206
#define TAG_METERING_MODE	0x9207
#define TAG_LIGHT_SOURCE	0x9208
#define TAG_FLASH		0x9209
#define TAG_FOCAL_LENGTH	0x920A
#define TAG_MAKERNOTE		0x927C

struct TagFormatAssoc {
	int tag;
	int type;
} IFD_tag_format[] = {
	{TAG_IMAGE_DESC,	FORMAT_STRING},
	{TAG_MAKE,		FORMAT_STRING},
	{TAG_MODEL,		FORMAT_STRING},
	{TAG_ORIENTATION,	FORMAT_USHORT},
	{TAG_RESOLUTION_X,	FORMAT_URATIONAL},
	{TAG_RESOLUTION_Y,	FORMAT_URATIONAL},
	{TAG_RESOLUTION_UNIT,	FORMAT_USHORT},
	{TAG_SOFTWARE,		FORMAT_STRING},
	{TAG_DATETIME,		FORMAT_STRING},
	{TAG_COPYRIGHT,		FORMAT_STRING},
	{TAG_EXIFOFFSET,	FORMAT_ULONG},
	{TAG_EXPOSURE_TIME,	FORMAT_URATIONAL},
	{TAG_FNUMBER,		FORMAT_URATIONAL},
	{TAG_EXPOSURE_PROGRAM,	FORMAT_USHORT},
	{TAG_ISO_SPEED,		FORMAT_USHORT},
	{TAG_DATETIME_ORIGINAL,	FORMAT_STRING},
	{TAG_DATETIME_DIGITIZED,FORMAT_STRING},
	{TAG_SHUTTER_SPEED,	FORMAT_SRATIONAL},
	{TAG_APERTURE,		FORMAT_URATIONAL},
	{TAG_EXPOSURE_BIAS,	FORMAT_SRATIONAL},
	{TAG_APERTURE_MAX,	FORMAT_URATIONAL},
	{TAG_SUBJECT_DISTANCE,	FORMAT_SRATIONAL},
	{TAG_METERING_MODE,	FORMAT_USHORT},
	{TAG_LIGHT_SOURCE,	FORMAT_USHORT},
	{TAG_FLASH,		FORMAT_USHORT},
	{TAG_FOCAL_LENGTH,	FORMAT_URATIONAL},
	{TAG_MAKERNOTE,		FORMAT_UNDEFINED}
};

struct TagValString {
	int tag_val;
	char *str;
};

struct TagValString tag_val_res_unit[] = {
	{EXIF_RESUNIT_NOUNIT,			""},
	{EXIF_RESUNIT_INCH,			N_("inch")},
	{EXIF_RESUNIT_CENTIMETER,		N_("cm")}
};

struct TagValString tag_val_exposure_program[] = {
	{EXIF_EXPOSURE_PROGRAM_UNKNOWN,		N_("unknown")},
	{EXIF_EXPOSURE_PROGRAM_MANUAL,		N_("manual")},
	{EXIF_EXPOSURE_PROGRAM_NORMAL,		N_("normal")},
	{EXIF_EXPOSURE_PROGRAM_APERTURE,	N_("aperture priority")},
	{EXIF_EXPOSURE_PROGRAM_SHUTTER,		N_("shutter priority")},
	{EXIF_EXPOSURE_PROGRAM_CREATIVE,	N_("creative")},
	{EXIF_EXPOSURE_PROGRAM_ACTION,		N_("action")},
	{EXIF_EXPOSURE_PROGRAM_PORTRAIT,	N_("portrait")},
	{EXIF_EXPOSURE_PROGRAM_LANDSCAPE,	N_("landscape")}
};

struct TagValString tag_val_metering_mode[] = {
	{EXIF_METERING_MODE_UNKNOWN,		N_("unknown")},
	{EXIF_METERING_MODE_AVERAGE,		N_("average")},
	{EXIF_METERING_MODE_CENTER_WEIGHTED,	N_("center weighted")},
	{EXIF_METERING_MODE_SPOT,		N_("spot")},
	{EXIF_METERING_MODE_MULTI_SPOT,		N_("multi-spot")},
	{EXIF_METERING_MODE_MULTI_SEGMENT,	N_("multi-segment")},
	{EXIF_METERING_MODE_PARTIAL,		N_("partial")},
	{EXIF_METERING_MODE_OTHER,		N_("other")}
};

struct TagValString tag_val_light_source[] = {
	{EXIF_LIGHT_SOURCE_UNKNOWN,		N_("unknown")},
	{EXIF_LIGHT_SOURCE_DAYLIGHT,		N_("daylight")},
	{EXIF_LIGHT_SOURCE_FLUORESCENT,		N_("fluorescent")},
	{EXIF_LIGHT_SOURCE_TUNGSTEN,		N_("tungsten")},
	{EXIF_LIGHT_SOURCE_FLASH,		N_("flash")},
	{EXIF_LIGHT_SOURCE_STANDARD_A,		N_("standard A")},
	{EXIF_LIGHT_SOURCE_STANDARD_B,		N_("standard B")},
	{EXIF_LIGHT_SOURCE_STANDARD_C,		N_("standard C")},
	{EXIF_LIGHT_SOURCE_D55,			"D55"},
	{EXIF_LIGHT_SOURCE_D65,			"D65"},
	{EXIF_LIGHT_SOURCE_D75,			"D75"},
	{EXIF_LIGHT_SOURCE_OTHER,		N_("OTHER")}
};

struct TagValString tag_val_flash[] = {
	{EXIF_FLASH_UNKNWON,			N_("unknown")},
	{EXIF_FLASH_NO,				N_("no")},
#if 0
	{EXIF_FLASH_YES,			N_("yes")},
	{EXIF_FLASH_RETURN_NOT_DETECTED,	N_("yes (strobe return light not detected)")},
	{EXIF_FLASH_RETURN_DETECTED,		N_("yes (strobe return light detected)")}
#else
	{EXIF_FLASH_YES,			N_("yes")}
#endif
};

typedef struct {
	char		byte_order[2];
	uint16_t	magic;
	uint32_t	IFD_offset;
} TIFFHeader;

typedef struct {
	uint16_t	tag;
	uint16_t	format;
	uint32_t	nb;
	uint32_t	data;
} IFDEntry;


static int parse_IFD_table(ExifData *exif, unsigned char *tiff, int offset,
			   int size, int byte_order);

/*
 *-------------------------------------------------------------------
 * file mmap
 *-------------------------------------------------------------------
 */

static gint map_file(const gchar *path, void **mapping, int *size)
{
	int fd;
	struct stat fs;

	if ((fd = open(path, O_RDONLY)) == -1)
		{
		perror(path);
		return -1;
		}

	if (fstat(fd, &fs) == -1)
		{
		perror(path);
		close(fd);
		return -1;
		}

	*size = fs.st_size;

	if ((*mapping = mmap(0, *size, PROT_READ|PROT_WRITE, MAP_PRIVATE, fd, 0)) == MAP_FAILED)
		{
		perror(path);
		close(fd);
		return -1;
		}

	close(fd);
	return 0;
}

static gint unmap_file(void *mapping, int size)
{
	if (munmap(mapping, size) == -1)
		{
		perror("munmap");
		return -1;
		}

	return 0;
}

/*
 *-------------------------------------------------------------------
 * byte size utils
 *-------------------------------------------------------------------
 */

static uint16_t get_int16(unsigned char *f, int bo)
{
	return bo == BYTE_ORDER_INTEL ?
		*f + (*(f+1)<<8) : ((*f)<<8) + *(f+1);
}

#if 0
/* not used ? */
static uint32_t get_int32(unsigned char *f, int bo)
{
	if (bo == BYTE_ORDER_INTEL)
		return get_int16(f, BYTE_ORDER_INTEL) + (get_int16(f+2, BYTE_ORDER_INTEL)<<16);
	else
		return (get_int16(f, BYTE_ORDER_MOTOROLA)<<16) + get_int16(f+2, BYTE_ORDER_MOTOROLA);
}
#endif

static uint16_t swab_int16(uint16_t n, int bo)
{
#if BYTE_ORDER == LITTLE_ENDIAN
	if (bo == BYTE_ORDER_MOTOROLA)
#else
	if (bo == BYTE_ORDER_INTEL)
#endif
		return n>>8 | n<<8 ;
	else
		return n;
}

static uint32_t swab_int32(uint32_t n, int bo)
{
#if BYTE_ORDER == LITTLE_ENDIAN
	if (bo == BYTE_ORDER_MOTOROLA)
#else
	if (bo == BYTE_ORDER_INTEL)
#endif
		return n<<24 | n>>24 | (n & 0xFF0000)>>8 | (n & 0xFF00)<<8;
	else
		return n;
}

/*
 *-------------------------------------------------------------------
 * marker utils
 *-------------------------------------------------------------------
 */


static int get_marker_size(unsigned char *f)
{
	/* Size is always in Motorola byte order */
	return get_int16(f+2, BYTE_ORDER_MOTOROLA);
}

static int goto_next_marker(unsigned char **f, int *size, int *marker)
{
	int marker_size = 2;

	*marker = MARKER_UNKNOWN;

	/* It is safe to access the marker and its size since we have checked
	 * the SOI and this function guaranties the whole next marker is
	 * available
	 */
	if (*(*f+1) != MARKER_SOI)
		{
		marker_size += get_marker_size(*f);
		}

	*size -= marker_size;

	/* size should be at least 4, so we can read the marker and its size
	 * and check data are actually available
	 */
	if (*size < 4) return -1;

	/* Jump to the next marker and be sure it begins with 0xFF
	 */
	*f += marker_size;
	if (**f != 0xFF) return -1;

	if (get_marker_size(*f)+2 > *size) return -1;

	*marker = *(*f+1);

	return 0;
}

/*
 *-------------------------------------------------------------------
 * IFD utils
 *-------------------------------------------------------------------
 */

static gchar *IFD_get_string(IFDEntry *ent, char *s, gint len)
{
	return g_strndup(s, len);
}

static long int IFD_get_integer(IFDEntry *ent, void *i, int byte_order)
{
	long int res=0;

	switch (ent->format)
		{
		case FORMAT_UBYTE:
		case FORMAT_SBYTE:
			res = *(char*)i;
			break;
		case FORMAT_USHORT:
		case FORMAT_SSHORT:
			res = swab_int16(*(uint16_t*)i, byte_order);
			break;
		case FORMAT_ULONG:
		case FORMAT_SLONG:
			res = swab_int32(*(uint32_t*)i, byte_order);
			break;
		}

	return res;
}

static void IFD_get_rational(IFDEntry *ent, void *i, int byte_order, ExifRational *n)
{
	n->num = swab_int32(*(uint32_t*)i, byte_order);
	n->den = swab_int32(*(uint32_t*)(i+sizeof(uint32_t)), byte_order);
}

static double IFD_get_double(IFDEntry *ent, void *i, int byte_order)
{
	ExifRational n;

	IFD_get_rational(ent, i, byte_order, &n);
	return (double)((int)n.num) / n.den;
}

static double IFD_get_unsigned_double(IFDEntry *ent, void *i, int byte_order)
{
	ExifRational n;

	IFD_get_rational(ent, i, byte_order, &n);
	return (double)(n.num) / n.den;
}

static uint16_t IFD_get_tag_format(uint16_t tag)
{
	static int len = sizeof(IFD_tag_format)/sizeof(struct TagFormatAssoc);
	int i = 0;

	while (i<len && IFD_tag_format[i].tag!=tag)
		{
		++i;
		}

	return (i == len ? FORMAT_UNDEFINED : IFD_tag_format[i].type);
}

static int parse_IFD_entry(ExifData *exif, unsigned char *tiff, int offset,
			   int size, int byte_order)
{
	IFDEntry *ent = (IFDEntry*)(tiff+offset);
	uint32_t swabed_data;
	void *data;
	int data_len;
	int std_format;

	ent->tag = swab_int16(ent->tag, byte_order);
	ent->format = swab_int16(ent->format, byte_order);
	ent->nb = swab_int32(ent->nb, byte_order);
	swabed_data = swab_int32(ent->data, byte_order);
#ifdef DEBUG_Q
	fprintf(stderr, "Tag: %x - ", ent->tag);
	fprintf(stderr, "Format: %x (%d bytes) - ", ent->format, IFD_format_size[ent->format-1]);
	fprintf(stderr, "Nb comp: %d - ", ent->nb);
	fprintf(stderr, "%s: %x\n",
		IFD_format_size[ent->format-1]*ent->nb > sizeof(ent->data) ?
		"Offset" : "Data", swabed_data);
#endif
	/* Check tag type. If it does not match, either the format is wrong,
	 * either it is a unknown tag; so it is not really an error.
	 */
	std_format = IFD_get_tag_format(ent->tag);
	if (std_format != ent->format)
		{
		/* Some cameras got mixed up signed/unsigned_rational
		 * eg KODAK DC4800 on object_distance tag
		 */
		if (((std_format != FORMAT_URATIONAL) && 
		     (std_format != FORMAT_SRATIONAL)) &&
		    ((std_format != FORMAT_SRATIONAL) || 
		     (std_format != FORMAT_URATIONAL)))
			{
#ifdef DEBUG_Q
			fprintf(stderr, "Tag %x: format (0x%x) should be (0x%x)\n", ent->tag, ent->format, IFD_get_tag_format(ent->tag));
#endif
			return 0;
			}
		}

	/* Where is the data, is it available?
	 */
	data_len = IFD_format_size[ent->format-1] * ent->nb; 
	if (data_len > sizeof(ent->data))
		{
		if (size < swabed_data+data_len)
			{
			fprintf(stderr, "Tag %x ERROR\n", ent->tag);
			return -1;
			}
		data = (void*)tiff + swabed_data;
		}
	else
		{
		data = (void*)(&(ent->data));
		}
	if (ent->tag>>12 == 0)
		{
		switch (ent->tag)
			{
			case TAG_IMAGE_DESC:
				if (!exif->image_desc)
					{
					exif->image_desc = IFD_get_string(ent, data, data_len);
					}
				break;
			case TAG_MAKE:
				if (!exif->make)
					{
					exif->make = IFD_get_string(ent, data, data_len);
					}
				break;
			case TAG_MODEL:
				if (!exif->model)
					{
					exif->model = IFD_get_string(ent, data, data_len);
					}
				break;
			case TAG_ORIENTATION:
				exif->orientation = IFD_get_integer(ent, data, byte_order);
				break;
			case TAG_RESOLUTION_X:
				IFD_get_rational(ent, data, byte_order, &(exif->resolution_x));
				break;
			case TAG_RESOLUTION_Y:
				IFD_get_rational(ent, data, byte_order, &(exif->resolution_y));
				break;
			case TAG_RESOLUTION_UNIT:
				exif->resolution_unit = IFD_get_integer(ent, data, byte_order);
				break;
			case TAG_SOFTWARE:
				if (!exif->software)
					{
					exif->software = IFD_get_string(ent, data, data_len);
					}
				break;
			case TAG_DATETIME:
				if (!exif->date)
					{
					exif->date = IFD_get_string(ent, data, data_len);
					}
				break;
			case TAG_COPYRIGHT:
				if (!exif->copyright)
					{
					exif->copyright = IFD_get_string(ent, data, data_len);
					}
				break;
			}
		}
	else
		{
		switch (ent->tag)
			{
			case TAG_EXIFOFFSET:
			  	if (ent->format == 4)
					parse_IFD_table(exif, tiff, swabed_data, size, byte_order);
			  	break;
			case TAG_EXPOSURE_TIME:
				IFD_get_rational(ent, data, byte_order, &exif->exposure_time);
				break;
			case TAG_FNUMBER:
				exif->fnumber = IFD_get_unsigned_double(ent, data, byte_order);
				break;
			case TAG_EXPOSURE_PROGRAM:
			  	exif->exposure_program = IFD_get_integer(ent, data, byte_order);
				break;
			case TAG_ISO_SPEED:
				exif->iso_sensivity = IFD_get_integer(ent, data, byte_order);
				break;
			case TAG_DATETIME_ORIGINAL:
				if (!exif->date_original)
					{
					exif->date_original = IFD_get_string(ent, data, data_len);
					}
				break;
			case TAG_DATETIME_DIGITIZED:
				if (!exif->date_digitized)
					{
					exif->date_digitized = IFD_get_string(ent, data, data_len);
					}
				break;
			case TAG_SHUTTER_SPEED:
				exif->shutter_speed = pow(2.0, IFD_get_double(ent, data, byte_order));
				/* Correct exposure time to avoid values like 1/91s (seen on Minolta DImage 7)
				 */
				if (exif->shutter_speed>1 && (int)exif->shutter_speed-((int)(exif->shutter_speed/10))*10 == 1)
					{
					--(exif->shutter_speed);
					}
				break;
			case TAG_APERTURE:
				exif->aperture = pow(sqrt(2), IFD_get_unsigned_double(ent, data, byte_order));
				break;
			case TAG_EXPOSURE_BIAS:
				exif->exposure_bias = IFD_get_double(ent, data, byte_order);
				break;
			case TAG_APERTURE_MAX:
				exif->max_aperture = pow(sqrt(2), IFD_get_unsigned_double(ent, data, byte_order));
				break;
			case TAG_SUBJECT_DISTANCE:
				exif->subject_distance = IFD_get_double(ent, data, byte_order);
				break;
			case TAG_METERING_MODE:
				exif->metering_mode = IFD_get_integer(ent, data, byte_order);
				break;
			case TAG_LIGHT_SOURCE:
				exif->light_source = IFD_get_integer(ent, data, byte_order);
				break;
			case TAG_FLASH:
				exif->flash = IFD_get_integer(ent, data, byte_order) & 0x1 ?
					EXIF_FLASH_YES : EXIF_FLASH_NO;
				break;
			case TAG_FOCAL_LENGTH:
				exif->focal_length = IFD_get_unsigned_double(ent, data, byte_order);
				break;
			}
		}

	return 0;
}

static int parse_IFD_table(ExifData *exif, unsigned char *tiff, int offset,
			   int size, int byte_order)
{
	int i, nb_entries;

	/* We should be able to read number of entries in IFD0) */
	if (size < offset+2) return -1;

	nb_entries = get_int16(tiff+offset, byte_order);

	/* Entries and next IFD offset must be readable */
	if (size < offset+nb_entries*12+4) return -1;

	for (i=0; i<nb_entries; ++i)
		{
		parse_IFD_entry(exif, tiff, offset+2+i*sizeof(IFDEntry), size, byte_order);
		}

	return 0;
}

/*
 *-------------------------------------------------------------------
 * tiff support
 *-------------------------------------------------------------------
 */

static int parse_TIFF(ExifData *exif, unsigned char *tiff, int size)
{
	int byte_order, offset=0;

	if (size < sizeof(TIFFHeader))
		{
		return -1;
		}

	if (strncmp(((TIFFHeader*)tiff)->byte_order, "II", 2) == 0)
		{
		byte_order = BYTE_ORDER_INTEL;
		}
	else if (strncmp(((TIFFHeader*)tiff)->byte_order, "MM", 2) == 0)
		{
		byte_order = BYTE_ORDER_MOTOROLA;
		}
	else
		{
		return -1;
		}

	if (swab_int16(((TIFFHeader*)tiff)->magic, byte_order) != 0x002A)
		{
		return -1;
		}

	offset = swab_int32(((TIFFHeader*)tiff)->IFD_offset, byte_order);

	return parse_IFD_table(exif, tiff, offset, size, byte_order);
}

/*
 *-------------------------------------------------------------------
 * jpeg support
 *-------------------------------------------------------------------
 */

static int parse_JPEG(ExifData *exif, unsigned char *f, int size)
{
	int marker, marker_size;

	if (size<2 || *f!=0xFF || *(f+1)!=MARKER_SOI)
		{
		return -2;
		}

	do {
		if (goto_next_marker(&f, &size, &marker) == -1)
			{
			break;
			}
	} while (marker != MARKER_APP1);

	if (marker != MARKER_APP1)
		{
		return -2;
		}

	marker_size = get_marker_size(f)-2;
		
	if (marker_size<6 || strncmp((char*)f+4, "Exif\0\0", 6)!=0)
		{
		return -2;
		}

	return parse_TIFF(exif, f+10, marker_size-6);
}

/*
 *-------------------------------------------------------------------
 * init blank data
 *-------------------------------------------------------------------
 */

static void exif_init_struct(ExifData *d)
{
	d->image_desc = NULL;
	d->make = NULL;
	d->model = NULL;
	d->orientation = EXIF_ORIENTATION_UNKNOWN;
	d->resolution_x.num = 72;
	d->resolution_x.den = 1;
	d->resolution_y.num = 72;
	d->resolution_y.den = 1;
	d->resolution_unit = EXIF_RESUNIT_NOUNIT;
	d->software = NULL;
	d->date = NULL;
	d->copyright = NULL;
	d->exposure_time = EXIF_RATIONAL_UNKNOWN;
	d->fnumber = EXIF_APERTURE_UNKNOWN;
	d->exposure_program = EXIF_EXPOSURE_PROGRAM_UNKNOWN;
	d->iso_sensivity = EXIF_ISO_SENSIVITY_UNKNOWN;
	d->date_original = NULL;
	d->date_digitized = NULL;
	d->aperture = EXIF_APERTURE_UNKNOWN;
	d->max_aperture = EXIF_APERTURE_UNKNOWN;
	d->exposure_bias = 0;
	d->subject_distance = 0;
	d->metering_mode = EXIF_METERING_MODE_UNKNOWN;
	d->light_source = EXIF_LIGHT_SOURCE_UNKNOWN;
	d->flash = EXIF_FLASH_UNKNWON;
	d->focal_length = EXIF_FOCAL_LENGTH_UNKNOWN;
}

/*
 *-------------------------------------------------------------------
 * the public stuff
 *-------------------------------------------------------------------
 */

void exif_free(ExifData *exif)
{
	if (!exif) return;

	g_free(exif->image_desc);
	g_free(exif->make);
	g_free(exif->model);
	g_free(exif->software);
	g_free(exif->date);
	g_free(exif->copyright);
	g_free(exif->date_original);
	g_free(exif->date_digitized);

	g_free(exif);
}

gchar *exif_rational_string(ExifRational *r)
{
	if (r->den > 1)
		{
		return g_strdup_printf("%d/%d", r->num, r->den);
		}

	return g_strdup_printf("%d", r->num);
}

gchar *exif_orientation_string(ExifData *exif)
{
	gchar *row;
	gchar *col;

	switch (exif->orientation)
		{
		default:
		case EXIF_ORIENTATION_UNKNOWN:
			row = _("unknown"); col = _("unknown");
			break;
		case EXIF_ORIENTATION_TOP_LEFT:
			row = _("top"); col = _("left");
			break;
		case EXIF_ORIENTATION_TOP_RIGHT:
			row = _("top"); col = _("right");
			break;
		case EXIF_ORIENTATION_BOTTOM_LEFT:
			row = _("bottom"); col = _("left");
			break;
		case EXIF_ORIENTATION_BOTTOM_RIGHT:
			row = _("bottom"); col = _("right");
			break;
		case EXIF_ORIENTATION_LEFT_TOP:
			row = _("left"); col = _("top");
			break;
		case EXIF_ORIENTATION_RIGHT_TOP:
			row = _("right"); col = _("top");
			break;
		case EXIF_ORIENTATION_RIGHT_BOTTOM:
			row = _("right"); col = _("bottom");
			break;
		case EXIF_ORIENTATION_LEFT_BOTTOM:
			row = _("left"); col = _("bottom");
			break;
		}

	return g_strdup_printf("%s, %s", row, col);
}

const gchar *exif_get_info_string(ExifData *exif, gint info)
{
	struct TagValString *table;
	int tag_val, size, i=0;

	switch(info)
		{
		case EXIF_RESUNIT:
			table = tag_val_res_unit;
			size = sizeof(tag_val_res_unit);
			tag_val = exif->resolution_unit;
			break;
		case EXIF_EXPOSURE_PROGRAM:
			table = tag_val_exposure_program;
			size = sizeof(tag_val_exposure_program);
			tag_val = exif->exposure_program;
			break;
		case EXIF_METERING_MODE:
			table = tag_val_metering_mode;
			size = sizeof(tag_val_metering_mode);
			tag_val = exif->metering_mode;
			break;
		case EXIF_LIGHT_SOURCE:
			table = tag_val_light_source;
			size = sizeof(tag_val_light_source);
			tag_val = exif->light_source;
			break;
		case EXIF_FLASH:
			table = tag_val_flash;
			size = sizeof(tag_val_flash);
			tag_val = exif->flash;
			break;
		default:
			return "";
		}

	size /= sizeof(struct TagValString);

	while (i < size && table[i].tag_val != tag_val)
		{
		++i;
		}

	return (i < size ? table[i].str : NULL);
}

#if 0
/* This remains for reference only, needs small fixup to work as debug */
void exif_print(ExifData *d)
{
	gchar *buf;

	printf("File:              %s\n", imd->image_path);
	printf("Camera:            %s %s (firmware %s)\n", d->make, d->model, d->software);
	printf("Date:              %s\n", d->date);
	printf("Date original:     %s\n", d->date_original);
	printf("Date digitized:    %s\n", d->date_digitized);
	printf("Copyright:         %s\n", d->copyright);

	buf = exif_orientation_string(d);
	printf("Orientation:       %s\n", buf);
	g_free(buf);

	printf("Resolution:        %d/%d, %d/%d %s\n",
		d->resolution_x.num, d->resolution_x.den,
		d->resolution_y.num, d->resolution_y.den,
		exif_get_info_string(imd, EXIF_RESUNIT));

	printf("Exposure program:  %s\n",
		exif_get_info_string(imd, EXIF_EXPOSURE_PROGRAM));

	printf("Shutter speed:     %s%.0f s\n",
	 	d->shutter_speed>1 ? "1/" : "",
	 	d->shutter_speed>1 ? floor(d->shutter_speed) : 1.0/d->shutter_speed);
	printf("Exposure time:     %d/%d s\n",
	       d->exposure_time.num, d->exposure_time.den);
	printf("Sensivity:         %d ISO\n", d->iso_sensivity);
	printf("Aperture:          f/%.1f (max: f/%.1f)\n", d->aperture, d->max_aperture);
	printf("Fnumber:             f/%.1f (max: f/%.1f)\n", d->fnumber, d->max_aperture);
	printf("Exposure bias:     %.1f\n", d->exposure_bias);
	printf("Focal length:      %.2f mm\n", d->focal_length);
	printf("Subject distance:  %.3f m\n", d->subject_distance);

	printf("Metering mode:     %s\n",
		exif_get_info_string(imd, EXIF_METERING_MODE));

	printf("Light source:      %s\n",
		exif_get_info_string(imd, EXIF_LIGHT_SOURCE));

	printf("Flash:             %s\n",
		exif_get_info_string(imd, EXIF_FLASH));

	printf("\n");
}
#endif

ExifData *exif_read(const gchar *path)
{
	ExifData *ed;
	unsigned char *f;
	int size, res;

	if (!path) return NULL;

	if (map_file(path, (void**)&f, &size) == -1)
		{
		return NULL;
		}

	ed = g_new0(ExifData, 1);
	exif_init_struct(ed);

	if ((res = parse_JPEG(ed, f, size)) == -2)
		{
		res = parse_TIFF(ed, f, size);
		}

	if (res != 0)
		{
		exif_free(ed);
		ed = NULL;
		}

	unmap_file(f, size);

	return ed;
}

