/*
 * GQview
 * (C) 2002 John Ellis
 *
 * Author: John Ellis
 *
 * This software is released under the GNU General Public License (GNU GPL).
 * Please read the included file COPYING for more information.
 * This software comes with no warranty of any kind, use at your own risk!
 */

#include "gqview.h"
#include "view_dir_list.h"

#include "dnd.h"
#include "dupe.h"
#include "filelist.h"
#include "layout.h"
#include "layout_image.h"
#include "layout_util.h"
#include "utilops.h"
#include "ui_clist_edit.h"
#include "ui_fileops.h"
#include "ui_menu.h"

#include <gdk/gdkkeysyms.h> /* for keyboard values */


#define VDLIST_PAD 4


/*
 *-----------------------------------------------------------------------------
 * misc
 *-----------------------------------------------------------------------------
 */

static gint vdlist_rename_row_cb(ClistEditData *ced, const gchar *old, const gchar *new, gpointer data)
{
	ViewDirList *vdl = data;
	gchar *old_path;
	gchar *new_path;
	gchar *base;

	old_path = g_strdup(vdlist_row_get_path(vdl, ced->row));
	if (!old_path) return FALSE;

	base = remove_level_from_path(old_path);
	new_path = concat_dir_and_file(base, new);
	g_free(base);

	if (rename(old_path, new_path) < 0)
		{
		gchar *buf;

		buf = g_strdup_printf(_("Failed to rename %s to %s."), old, new);
		warning_dialog("Rename failed", buf);
		g_free(buf);
		}
	else
		{
		if (vdl->layout && strcmp(vdl->path, old_path) == 0)
			{
			layout_set_path(vdl->layout, new_path);
			}
		else
			{
			vdlist_refresh(vdl);
			}
		}

	g_free(old_path);
	g_free(new_path);
	return FALSE;
}

static void vdlist_rename_by_row(ViewDirList *vdl, gint row)
{
	if (row < 0) return;

	clist_edit_by_row(GTK_CLIST(vdl->clist), row, 0,
			  vdlist_rename_row_cb, vdl);
}

static gint vdlist_row_by_path(ViewDirList *vdl, const gchar *path)
{
	GList *work;
	gint n;

	if (!path) return -1;

	n = 0;
	work = vdl->list;
	while (work)
		{
		FileData *fd = work->data;
		if (strcmp(fd->path, path) == 0) return n;
		work = work->next;
		n++;
		}

	return -1;
}

/*
 *-----------------------------------------------------------------------------
 * drop menu (from dnd)
 *-----------------------------------------------------------------------------
 */

static void vdlist_drop_menu_copy_cb(GtkWidget *widget, gpointer data)
{
	ViewDirList *vdl = data;
	const gchar *path;
	GList *list;

	path = vdlist_row_get_path(vdl, vdl->drop_row);
	list = vdl->drop_list;

	vdl->drop_list = NULL;

	file_util_copy_simple(list, path);
}

static void vdlist_drop_menu_move_cb(GtkWidget *widget, gpointer data)
{
	ViewDirList *vdl = data;
	const gchar *path;
	GList *list;

	path = vdlist_row_get_path(vdl, vdl->drop_row);
	list = vdl->drop_list;

	vdl->drop_list = NULL;

	file_util_move_simple(list, path);
}

static GtkWidget *vdlist_drop_menu(ViewDirList *vdl, gint active)
{
	GtkWidget *menu;

	menu = popup_menu_short_lived();

	menu_item_add_stock_sensitive(menu, _("Copy"), GTK_STOCK_COPY, active,
				      GTK_SIGNAL_FUNC(vdlist_drop_menu_copy_cb), vdl);
	menu_item_add_sensitive(menu, _("Move"), active, GTK_SIGNAL_FUNC(vdlist_drop_menu_move_cb), vdl);

	menu_item_add_divider(menu);
	menu_item_add_stock(menu, _("Cancel"), GTK_STOCK_CANCEL, NULL, vdl);

	return menu;
}

/*
 *-----------------------------------------------------------------------------
 * pop-up menu
 *-----------------------------------------------------------------------------
 */ 

static void vdlist_pop_menu_up_cb(GtkWidget *widget, gpointer data)
{
	ViewDirList *vdl = data;
	gchar *path;

	if (!vdl->path || strcmp(vdl->path, "/") == 0) return;
	path = remove_level_from_path(vdl->path);

	if (vdl->select_func)
		{
		vdl->select_func(vdl, path, vdl->select_data);
		}

	g_free(path);
}

static void vdlist_pop_menu_slide_cb(GtkWidget *widget, gpointer data)
{
	ViewDirList *vdl = data;
	gchar *path;

	if (!vdl->layout) return;

	path = g_strdup(vdlist_row_get_path(vdl, vdl->click_row));
	if (!path) return;

	layout_set_path(vdl->layout, path);
	layout_select_none(vdl->layout);
	layout_image_slideshow_stop(vdl->layout);
	layout_image_slideshow_start(vdl->layout);

	g_free(path);
}

static void vdlist_pop_menu_slide_rec_cb(GtkWidget *widget, gpointer data)
{
	ViewDirList *vdl = data;
	gchar *path;
	GList *list;

	if (!vdl->layout) return;

	path = g_strdup(vdlist_row_get_path(vdl, vdl->click_row));
	if (!path) return;

	list = path_list_recursive(path);

	layout_image_slideshow_stop(vdl->layout);
	layout_image_slideshow_start_from_list(vdl->layout, list);

	g_free(path);
}

static void vdlist_pop_menu_dupe(ViewDirList *vdl, gint recursive)
{
	DupeWindow *dw;
	const gchar *path;
	GList *list = NULL;

	path = vdlist_row_get_path(vdl, vdl->click_row);
	if (!path) return;

	if (recursive)
		{
		list = g_list_append(list, g_strdup(path));
		}
	else
		{
		path_list(path, &list, NULL);
		list = path_list_filter(list, FALSE);
		}

	dw = dupe_window_new(DUPE_MATCH_NAME);
	dupe_window_add_files(dw, list, recursive);

	path_list_free(list);
}

static void vdlist_pop_menu_dupe_cb(GtkWidget *widget, gpointer data)
{
	ViewDirList *vdl = data;
	vdlist_pop_menu_dupe(vdl, FALSE);
}

static void vdlist_pop_menu_dupe_rec_cb(GtkWidget *widget, gpointer data)
{
	ViewDirList *vdl = data;
	vdlist_pop_menu_dupe(vdl, TRUE);
}

static void vdlist_pop_menu_new_cb(GtkWidget *widget, gpointer data)
{
	ViewDirList *vdl = data;
	gchar *new_path;
	gchar *buf;

	if (!vdl->path) return;

	buf = concat_dir_and_file(vdl->path, _("new_folder"));
	new_path = unique_filename(buf, NULL, NULL, FALSE);
	g_free(buf);
	if (!new_path) return;

	if (mkdir(new_path, 0755) < 0)
		{
		gchar *text;

		text = g_strdup_printf(_("Unable to create directory:\n%s"), new_path);
		file_util_warning_dialog(_("Error creating directory"), text);
		g_free(text);
		}
	else
		{
		gint new_row;

		vdlist_refresh(vdl);
		new_row = vdlist_row_by_path(vdl, new_path);

		vdlist_rename_by_row(vdl, new_row);
		}

	g_free(new_path);
}

static void vdlist_pop_menu_rename_cb(GtkWidget *widget, gpointer data)
{
	ViewDirList *vdl = data;

	vdlist_rename_by_row(vdl, vdl->click_row);
}

static void vdlist_pop_menu_tree_cb(GtkWidget *widget, gpointer data)
{
	ViewDirList *vdl = data;

	if (vdl->layout) layout_views_set(vdl->layout, TRUE, vdl->layout->icon_view);
}

static void vdlist_pop_menu_refresh_cb(GtkWidget *widget, gpointer data)
{
	ViewDirList *vdl = data;

	if (vdl->layout) layout_refresh(vdl->layout);
}

static GtkWidget *vdlist_pop_menu(ViewDirList *vdl, gint row)
{
	GtkWidget *menu;
	gint active;

	active = (row >= 0);

	menu = popup_menu_short_lived();

	menu_item_add_stock_sensitive(menu, _("Up to parent"), GTK_STOCK_GO_UP,
				      (vdl->path && strcmp(vdl->path, "/") != 0),
				      GTK_SIGNAL_FUNC(vdlist_pop_menu_up_cb), vdl);

	menu_item_add_divider(menu);
	menu_item_add_sensitive(menu, _("Slideshow"), active,
				GTK_SIGNAL_FUNC(vdlist_pop_menu_slide_cb), vdl);
	menu_item_add_sensitive(menu, _("Slideshow recursive"), active,
				GTK_SIGNAL_FUNC(vdlist_pop_menu_slide_rec_cb), vdl);

	menu_item_add_divider(menu);
	menu_item_add_stock_sensitive(menu, _("Find duplicates..."), GTK_STOCK_FIND, active,
				      GTK_SIGNAL_FUNC(vdlist_pop_menu_dupe_cb), vdl);
	menu_item_add_stock_sensitive(menu, _("Find duplicates recursive..."), GTK_STOCK_FIND, active,
				      GTK_SIGNAL_FUNC(vdlist_pop_menu_dupe_rec_cb), vdl);

	menu_item_add_divider(menu);

	/* check using . (always row 0) */
	active = (vdl->path && access(vdl->path , W_OK | X_OK) == 0);
	menu_item_add_sensitive(menu, _("New folder..."), active,
				GTK_SIGNAL_FUNC(vdlist_pop_menu_new_cb), vdl);

	/* ignore .. and . */
	active = (row >= 2 && active &&
		  access(vdlist_row_get_path(vdl, vdl->click_row), W_OK | X_OK) == 0);
	menu_item_add_sensitive(menu, _("Rename..."), active,
				GTK_SIGNAL_FUNC(vdlist_pop_menu_rename_cb), vdl);

	menu_item_add_divider(menu);
	menu_item_add_check(menu, _("View as tree"), FALSE,
			    GTK_SIGNAL_FUNC(vdlist_pop_menu_tree_cb), vdl);
	menu_item_add_stock(menu, _("Refresh"), GTK_STOCK_REFRESH,
			    GTK_SIGNAL_FUNC(vdlist_pop_menu_refresh_cb), vdl);

	return menu;
}

static void vdlist_popup_destroy_cb(GtkWidget *widget, gpointer data)
{
	ViewDirList *vdl = data;
	clist_edit_set_highlight(vdl->clist, vdl->click_row, FALSE);
	vdl->click_row = -1;
	vdl->popup = NULL;

	path_list_free(vdl->drop_list);
	vdl->drop_list = NULL;
	vdl->drop_row = -1;
}

/*
 *-----------------------------------------------------------------------------
 * dnd
 *-----------------------------------------------------------------------------
 */

static GtkTargetEntry vdlist_dnd_drop_types[] = {
	{ "text/uri-list", 0, TARGET_URI_LIST }
};
static gint vdlist_dnd_drop_types_count = 1;


static void vdlist_set_prelight(ViewDirList *vdl, gint row, gint prelight)
{
	if (row < 0 || row == vdl->click_row) return;

	if (prelight)
		{
		gtk_clist_set_background(GTK_CLIST(vdl->clist), row,
			&GTK_WIDGET(vdl->clist)->style->bg[GTK_STATE_PRELIGHT]);
		gtk_clist_set_foreground(GTK_CLIST(vdl->clist), row,
			&GTK_WIDGET(vdl->clist)->style->fg[GTK_STATE_PRELIGHT]);
		}
	else
		{
		gtk_clist_set_background(GTK_CLIST(vdl->clist), row, NULL);
		gtk_clist_set_foreground(GTK_CLIST(vdl->clist), row, NULL);
		}
}

static void vdlist_dnd_get(GtkWidget *widget, GdkDragContext *context,
			   GtkSelectionData *selection_data, guint info,
			   guint time, gpointer data)
{
	ViewDirList *vdl = data;

	if (vdl->click_row >= 0)
		{
		const gchar *path;
		gchar *text = NULL;

		path = vdlist_row_get_path(vdl, vdl->click_row);

		switch (info)
			{
			case TARGET_URI_LIST:
				text = g_strconcat("file:", path, "\r\n", NULL);
				break;
			case TARGET_TEXT_PLAIN:
				text = g_strdup(path);
				break;
			}
		if (text)
			{
			gtk_selection_data_set (selection_data, selection_data->target,
					8, text, strlen(text));
			g_free(text);
			}
		}
}

static void vdlist_dnd_begin(GtkWidget *widget, GdkDragContext *context, gpointer data)
{
	ViewDirList *vdl = data;

	clist_edit_set_highlight(vdl->clist, vdl->click_row, TRUE);
}

static void vdlist_dnd_end(GtkWidget *widget, GdkDragContext *context, gpointer data)
{
	ViewDirList *vdl = data;

	clist_edit_set_highlight(vdl->clist, vdl->click_row, FALSE);

	if (context->action == GDK_ACTION_MOVE)
		{
		vdlist_refresh(vdl);
		}
}

static void vdlist_dnd_drop_receive(GtkWidget *widget,
				    GdkDragContext *context, gint x, gint y,
				    GtkSelectionData *selection_data, guint info,
				    guint time, gpointer data)
{
	ViewDirList *vdl = data;
	gint row = -1;

	gtk_clist_get_selection_info(GTK_CLIST(vdl->clist), x, y, &row, NULL);

	if (row == -1) return;

	if (info == TARGET_URI_LIST)
		{
		GList *list;
		gint active;

		list = get_uri_file_list(selection_data->data);
		if (!list) return;

		active = (access(vdlist_row_get_path(vdl, row), W_OK | X_OK) == 0);

		vdl->popup = vdlist_drop_menu(vdl, active);
		gtk_menu_popup(GTK_MENU(vdl->popup), NULL, NULL, NULL, NULL, 0, time);
		gtk_signal_connect(GTK_OBJECT(vdl->popup), "destroy",
				   GTK_SIGNAL_FUNC(vdlist_popup_destroy_cb), vdl);

		vdl->drop_row = row;
		vdl->drop_list = list;
		}
}

static void vdlist_dnd_drop_scroll_cancel(ViewDirList *vdl)
{
	if (vdl->drop_scroll_id != -1) gtk_timeout_remove(vdl->drop_scroll_id);
	vdl->drop_scroll_id = -1;
}

static gint vdlist_dnd_drop_scroll_cb(gpointer data)
{
	ViewDirList *vdl = data;

	if (vdl->drop_row == -1 || vdl->drop_list)
		{
		vdl->drop_scroll_id = -1;
		return FALSE;
		}
	if (vdl->drop_row < 1 || vdl->drop_row >= GTK_CLIST(vdl->clist)->rows - 1) return TRUE;

	vdlist_set_prelight(vdl, vdl->drop_row, FALSE);
	if (gtk_clist_row_is_visible(GTK_CLIST(vdl->clist), vdl->drop_row - 1) != GTK_VISIBILITY_FULL)
		{
		gtk_clist_moveto(GTK_CLIST(vdl->clist), vdl->drop_row - 1, 0, 0.0, 0.0);
		if (vdl->drop_row > 0)
			{
			vdl->drop_row--;
			}
		}
	else if (gtk_clist_row_is_visible(GTK_CLIST(vdl->clist), vdl->drop_row + 1) != GTK_VISIBILITY_FULL)
		{
		gtk_clist_moveto(GTK_CLIST(vdl->clist), vdl->drop_row + 1, 0, 1.0, 0.0);
		if (vdl->drop_row < GTK_CLIST(vdl->clist)->rows - 1)
			{
			vdl->drop_row++;
			}
		}
	vdlist_set_prelight(vdl, vdl->drop_row, TRUE);

	return TRUE;
}

static gint vdlist_dnd_drop_motion(GtkWidget *widget, GdkDragContext *context,
				   gint x, gint y, guint time, gpointer data)
{
        ViewDirList *vdl = data;
	gint row = -1;

	gtk_clist_get_selection_info(GTK_CLIST(vdl->clist), x, y, &row, NULL);

	if (gtk_drag_get_source_widget(context) == vdl->clist)
		{
		/* from same window */
		gdk_drag_status(context, 0, time);
		}
	else
		{
		gdk_drag_status(context, context->suggested_action, time);
		}

	if (row != vdl->drop_row)
		{
		vdlist_set_prelight(vdl, vdl->drop_row, FALSE);
		vdlist_set_prelight(vdl, row, TRUE);
		}

	/* delayed drop scrolling */
	if (vdl->drop_row == -1 && row != -1 && vdl->drop_scroll_id == -1)
		{
		vdl->drop_scroll_id = gtk_timeout_add(100, vdlist_dnd_drop_scroll_cb, vdl);
		}
	vdl->drop_row = row;

	return FALSE;
}

static void vdlist_dnd_drop_leave(GtkWidget *widget, GdkDragContext *context, guint time, gpointer data)
{
	ViewDirList *vdl = data;

	if (vdl->drop_row != vdl->click_row) vdlist_set_prelight(vdl, vdl->drop_row, FALSE);

	vdl->drop_row = -1;
}

static void vdlist_dnd_init(ViewDirList *vdl)
{
	gtk_drag_source_set(vdl->clist, GDK_BUTTON2_MASK,
			    dnd_file_drag_types, dnd_file_drag_types_count,
			    GDK_ACTION_COPY | GDK_ACTION_MOVE | GDK_ACTION_LINK);
	gtk_signal_connect(GTK_OBJECT(vdl->clist), "drag_data_get",
			   GTK_SIGNAL_FUNC(vdlist_dnd_get), vdl);
	gtk_signal_connect(GTK_OBJECT(vdl->clist), "drag_begin",
			   GTK_SIGNAL_FUNC(vdlist_dnd_begin), vdl);
	gtk_signal_connect(GTK_OBJECT(vdl->clist), "drag_end",
			   GTK_SIGNAL_FUNC(vdlist_dnd_end), vdl);

	gtk_drag_dest_set(vdl->clist,
			GTK_DEST_DEFAULT_MOTION | GTK_DEST_DEFAULT_DROP,
			vdlist_dnd_drop_types, vdlist_dnd_drop_types_count,
			GDK_ACTION_MOVE | GDK_ACTION_COPY);
	gtk_signal_connect(GTK_OBJECT(vdl->clist), "drag_data_received",
			   GTK_SIGNAL_FUNC(vdlist_dnd_drop_receive), vdl);
	gtk_signal_connect(GTK_OBJECT(vdl->clist), "drag_motion",
			GTK_SIGNAL_FUNC(vdlist_dnd_drop_motion), vdl);
	gtk_signal_connect(GTK_OBJECT(vdl->clist), "drag_leave",
			GTK_SIGNAL_FUNC(vdlist_dnd_drop_leave), vdl);
}

/*
 *-----------------------------------------------------------------------------
 * main
 *-----------------------------------------------------------------------------
 */ 

const gchar *vdlist_row_get_path(ViewDirList *vdl, gint row)
{
	FileData *fd;

	fd = g_list_nth_data(vdl->list, row);

	if (fd) return fd->path;

	return NULL;
}

static void vdlist_populate(ViewDirList *vdl)
{
	GtkCList *clist;
	GdkFont *font;
	GList *work;
	gint width = 0;

	clist = GTK_CLIST(vdl->clist);
	font = gtk_style_get_font(GTK_WIDGET(clist)->style);

	gtk_clist_freeze(clist);
	gtk_clist_clear(clist);

	work = vdl->list;
	while (work)
		{
		FileData *fd;
		gint row;
		gchar *buf[2];
		gint tmp_width;
		GdkPixmap *pixmap;
		GdkBitmap *mask;

		fd = work->data;

		buf[0] = (gchar *)fd->name;
		buf[1] = NULL;

		row = gtk_clist_append(clist, buf);

		if (access(fd->path, R_OK) == 0)
			{
			pixmap = vdl->pf->close_pixmap;
			mask = vdl->pf->close_mask;
			}
		else
			{
			pixmap = vdl->pf->deny_pixmap;
			mask = vdl->pf->deny_mask;
			}

		gtk_clist_set_pixtext(clist, row, 0, fd->name, VDLIST_PAD,
				      pixmap, mask);
		gtk_clist_set_row_data(clist, row, fd);

		tmp_width = gdk_string_width(font, buf[0]);
		if (tmp_width > width) width = tmp_width;

		work = work->next;
		}

	gtk_clist_set_column_width(clist, 0, width);
	gtk_clist_thaw(clist);

	vdl->click_row = -1;
}

gint vdlist_set_path(ViewDirList *vdl, const gchar *path)
{
	gint ret;
	FileData *fd;
	gchar *old_path = NULL;

	if (!path) return FALSE;
	if (vdl->path && strcmp(path, vdl->path) == 0) return TRUE;

	if (vdl->path)
		{
		gchar *base;

		base = remove_level_from_path(vdl->path);
		if (strcmp(base, path) == 0)
			{
			old_path = g_strdup(filename_from_path(vdl->path));
			}
		g_free(base);
		}

	g_free(vdl->path);
	vdl->path = g_strdup(path);

	filelist_free(vdl->list);
	vdl->list = NULL;

	ret = filelist_read(vdl->path, NULL, &vdl->list);

	vdl->list = filelist_sort(vdl->list, SORT_NAME, TRUE);

	/* add . and .. */

	fd = g_new0(FileData, 1);
	fd->path = remove_level_from_path(vdl->path);
	fd->name = "..";
	vdl->list = g_list_prepend(vdl->list, fd);

	fd = g_new0(FileData, 1);
	fd->path = g_strdup(vdl->path);
	fd->name = ".";
	vdl->list = g_list_prepend(vdl->list, fd);

	vdlist_populate(vdl);

	if (old_path)
		{
		/* scroll to make last path visible */
		gint row;
		gint found;
		GList *work;

		row = -1;
		found = FALSE;
		work = vdl->list;
		while (work && !found)
			{
			FileData *fd;

			fd = work->data;
			work = work->next;

			row++;
			if (strcmp(old_path, fd->name) == 0) found = TRUE;
			}

		if (found) gtk_clist_moveto(GTK_CLIST(vdl->clist), row, 0, 0.5, 0.0);

		g_free(old_path);
		}

	return ret;
}

void vdlist_refresh(ViewDirList *vdl)
{
	gchar *path;

	path = g_strdup(vdl->path);
	vdl->path = NULL;
	vdlist_set_path(vdl, path);
	g_free(path);
}

static void vdlist_menu_position_cb(GtkMenu *menu, gint *x, gint *y, gboolean *push_in, gpointer data)
{
	ViewDirList *vdl = data;
	GtkCList *clist;

	clist = GTK_CLIST(vdl->clist);
	gdk_window_get_origin(clist->clist_window, x, y);

	if (vdl->click_row >= 0)
		{
		*x += clist->column[0].area.x + clist->hoffset;
		*y += (clist->row_height + 1) * (vdl->click_row + 1) - 1 + clist->voffset;
		}

	popup_menu_position_clamp(menu, x, y);
}

static gint vdlist_press_key_cb(GtkWidget *widget, GdkEventKey *event, gpointer data)
{
	ViewDirList *vdl = data;
	GtkCList *clist;
	gint row;
	
	if (event->keyval != GDK_Menu) return FALSE;

	clist = GTK_CLIST(vdl->clist);
	row = clist->focus_row;
	if (row >= 0 && row < clist->rows)
		{
		vdl->click_row = row;
		}
	else
		{
		vdl->click_row = -1;
		}

	clist_edit_set_highlight(vdl->clist, vdl->click_row, TRUE);

	vdl->popup = vdlist_pop_menu(vdl, vdl->click_row);
	gtk_signal_connect(GTK_OBJECT(vdl->popup), "destroy",
			   GTK_SIGNAL_FUNC(vdlist_popup_destroy_cb), vdl);

	gtk_menu_popup(GTK_MENU(vdl->popup), NULL, NULL, vdlist_menu_position_cb, vdl, 0, GDK_CURRENT_TIME);

	return TRUE;
}

static gint vdlist_press_cb(GtkWidget *widget, GdkEventButton *bevent, gpointer data)
{
	ViewDirList *vdl = data;
	gint row = -1;
	gint col = -1;

	gtk_clist_get_selection_info(GTK_CLIST(widget), bevent->x, bevent->y, &row, &col);

	vdl->click_row = row;

	if (bevent->button == 3)
		{

		clist_edit_set_highlight(vdl->clist, row, TRUE);

		vdl->popup = vdlist_pop_menu(vdl, row);

		gtk_signal_connect(GTK_OBJECT(vdl->popup), "destroy",
				   GTK_SIGNAL_FUNC(vdlist_popup_destroy_cb), vdl);

		gtk_menu_popup(GTK_MENU(vdl->popup), NULL, NULL, NULL, NULL,
			       bevent->button, bevent->time);
		}
	return FALSE;
}

static void vdlist_select_cb(GtkWidget *widget, gint row, gint col,
			     GdkEvent *event, gpointer data)
{
	ViewDirList *vdl = data;
	gchar *path;

	path = g_strdup(vdlist_row_get_path(vdl, row));
	if (vdl->select_func)
		{
		vdl->select_func(vdl, path, vdl->select_data);
		}
	g_free(path);
}

static void vdlist_destroy_cb(GtkWidget *widget, gpointer data)
{
	ViewDirList *vdl = data;

	if (vdl->popup)
		{
		gtk_signal_disconnect_by_data(GTK_OBJECT(vdl->popup), vdl);
		gtk_widget_destroy(vdl->popup);
		}

	vdlist_dnd_drop_scroll_cancel(vdl);

	path_list_free(vdl->drop_list);

	folder_icons_free(vdl->pf);

	g_free(vdl->path);
	filelist_free(vdl->list);
	g_free(vdl);
}

ViewDirList *vdlist_new(const gchar *path)
{
	ViewDirList *vdl;

	vdl = g_new0(ViewDirList, 1);

	vdl->path = NULL;
	vdl->list = NULL;
	vdl->click_row = -1;

	vdl->drop_row = -1;
	vdl->drop_list = NULL;

	vdl->drop_scroll_id = -1;

	vdl->popup = NULL;

	vdl->widget = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(vdl->widget),
				       GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
	gtk_signal_connect(GTK_OBJECT(vdl->widget), "destroy",
			   GTK_SIGNAL_FUNC(vdlist_destroy_cb), vdl);

	vdl->clist = gtk_clist_new(1);
	gtk_signal_connect(GTK_OBJECT(vdl->clist), "button_press_event",
			   GTK_SIGNAL_FUNC(vdlist_press_cb), vdl);
	gtk_signal_connect(GTK_OBJECT(vdl->clist), "key_press_event",
			   GTK_SIGNAL_FUNC(vdlist_press_key_cb), vdl);
	gtk_signal_connect(GTK_OBJECT(vdl->clist), "select_row",
			   GTK_SIGNAL_FUNC(vdlist_select_cb), vdl);
	gtk_clist_column_titles_passive(GTK_CLIST(vdl->clist));
	gtk_container_add(GTK_CONTAINER(vdl->widget), vdl->clist);
	gtk_widget_show(vdl->clist);

	vdl->pf = folder_icons_new();

	vdlist_dnd_init(vdl);

	if (path) vdlist_set_path(vdl, path);

	return vdl;
}

void vdlist_set_select_func(ViewDirList *vdl,
			    void (*func)(ViewDirList *vdl, const gchar *path, gpointer data), gpointer data)
{
	vdl->select_func = func;
	vdl->select_data = data;
}

void vdlist_set_layout(ViewDirList *vdl, LayoutWindow *layout)
{
	vdl->layout = layout;
}

